#!/usr/bin/env python
""" Tools to manage array propery objects.
"""
from copy import deepcopy

import numpy as np

from .property import Property, defaults_decorator

class Array(Property):
    """Property sub-class numpy arrays
    """
    defaults = deepcopy(Property.defaults)
    defaults['dtype'] = (float, 'Data type of array element')
    defaults['default'] = (np.nan, 'Default value')

    @defaults_decorator(defaults)
    def __init__(self, **kwargs):
        super(Array, self).__init__(**kwargs)

    def _cast_type(self, value, obj=None):
        """Hook took override type casting"""
        return np.array(value).astype(self.dtype)



#!/usr/bin/env python
""" Tools to manage Property object that can be used as fit parameters.
"""
from collections.abc import Mapping
import numpy as np

from .array import Array
from .configurable import Configurable
from .utils import is_none

class ParamHolder(Configurable):
    """Wrapper around a data value

    This includes value, bounds, error estimates and fixed/free status
    (i.e., for fitting)

    """
    value = Array(help='Parameter Value')
    errors = Array(help='Parameter Uncertainties')
    bounds = Array(help='Parameter Bounds')
    scale = Array(default=1., help='Paramter Scale Factor')
    free = Array(dtype=bool, default=False, help='Free/Fixed Flag')

    def __init__(self, *args, **kwargs):
        """Constructor"""
        kwcopy = kwargs.copy()
        if args: #pragma: no cover
            #if 'value' in kwcopy:
            #    raise ValueError("value keyword provided in addition to arguments")
            kwcopy['value'] = args
        if is_none(kwargs.get('value', None)):
            kwargs.pop('value', None)                       
        super(ParamHolder, self).__init__(**kwargs)
        self.check_bounds()

    def __str__(self, indent=0):
        """Return self as a string"""
        return '{0:>{2}}{1}'.format('', '', indent) + "{_value} += {_errors} [{_bounds}] <{_scale}> {_free}".format(**self.__dict__)

    def update(self, *args, **kwargs):
        """Update the parameter"""
        kwcopy = kwargs.copy()
        if args:
            if 'value' in kwcopy:
                raise ValueError("value keyword provided in addition to arguments")
            if len(args) > 1:
                raise ValueError("Only 1 argument allowed")
            if isinstance(args[0], Mapping):
                kwcopy.update(**args[0])
            else:
                kwcopy['value'] = args[0]
        super(ParamHolder, self).update(**kwcopy)
        self.check_bounds()

    def check_bounds(self):
        """Hook for bounds-checking, invoked during assignment.

        raises ValueError if value is outside of bounds.
        does nothing if bounds is set to None.
        """
        if np.isnan(self.value).all():
            return
        if np.isnan(self.bounds).all():
            return
        if np.bitwise_or(self.value < self.bounds[0], self.value > self.bounds[-1]).any(): #pylint: disable=unsubscriptable-object
            raise ValueError("Value outside bounds: %.s [%s,%s]" % (self.value, self.bounds[0], self.bounds[-1])) #pylint: disable=unsubscriptable-object

    def __call__(self):
        """Return the product of the value and the scale"""
        return self.value * self.scale



#!/usr/bin/env python
""" Tools wrap functions with jax methods for automatic differentiation
"""
import numpy as np

from collections.abc import Iterable

from jax import grad, jacfwd, jacrev
import jax.numpy as jnp


def check_inputs(x, inputs):
    """Convert inputs to arrays"""
    ret = []
    if x is None:
        x = np.nan
    elif not isinstance(x, (Iterable, int, float)):
        raise TypeError("Non-numeric value %s passed in Physics" % (str(x)))

    x = np.array(x).astype(np.float)
    ret.append(x)
    if not inputs:
        return ret

    for inp in inputs:
        if callable(inp):
            ret.append(inp(x).astype(np.float))
        elif inp is None:
            ret.append(np.full(x.shape, np.nan))
        elif isinstance(inp, np.ndarray):
            ret.append(inp)
        elif isinstance(inp, (Iterable, int, float)):
            ret.append(np.array(inp).astype(np.float))
        else:
            raise TypeError("Non-numeric value %s passed in Physics" % (str(x)))
    return ret


def jax_hess(ifunc, argnums):
    """Return the hessian of a function

    Parameter
    ---------
    ifunc : `function`
        The inner function being wrapped
    argnum : `int` or `tuple`
        Indices of the arguments to get derivatvies for

    Return
    ------
    hess : `array`
        The Hessian matrix
    """
    return jacfwd(jacrev(ifunc, argnums), argnums)

def jax_wrapper(wfunc, ifunc, getargs, varidx, *args, **kwds):
    """Wrap a function with a jax function

    Parameter
    ---------
    wfunc : `function`
        The `jax` function to wrap with
    ifunc : `function`
        The inner function being wrapped
    getargs : `function`
        A function that formats the arguments to the ifunc
    varidx : `dict`
        Dictionary mapping variable names to indices

    Keywords
    --------
    argnums : `int` or `tuple`
        Indices of the arguments to get derivatives for
    argnames : `list`
        Names of the arguments to get derivatives for

    Return
    ------
    func : `function`
        The wrapped function

    Notes
    -----
    Get args should return a 2-D array, each row consisting a set of argument to the inner function

    args and kwds are passes to the getargs function
    """
    argnums = kwds.get('argnums', 0)
    argnames = kwds.get('argnames', None)
    if argnames is not None:
        argnums = [ varidx[argname] for argname in argnames ]
    args_ = getargs(*args, **kwds)
    wrapped = wfunc(ifunc, argnums)
    return jnp.array([wrapped(*args__) for args__ in args_])


class Function:
    """Decorator class to functions to add gradient, hessians to a functin

    Examples::

        class MyClass(Model):

            var1 = Parameter(default=0.3, help="A variable")
            var2 = Parameter(default=1.2, help="Another variable")

            @Function
            def product(extra_var, var1, var2):
                return extra_var*var1*var2

        # Make an object
        my_obj = MyClass()

        # Get the product, note that extra_var must be provied, but var1 and var2 are taken from the class
        my_obj.product(1.)

        # Get the product, overrided var1
        my_obj.product(1., var1=0.4)

        # Get the gradient for variable 0 (i.e., extra_var), evaluated at extra_var = 1.
        my_obj.product.grad(1.)

        # Get the gradient for variables 0,1,2, evaluated at extra_var = 1.
        my_obj.product.grad(1., argnums=[0,1,2])

        # Get the gradient for var1, var2, evaluated at extra_var = 1.
        my_obj.product.grad(1., argnames=["var1", "var2"])

        # Get the gradient for variable 0, evaluated at several points
        my_obj.product.grad([1.,1.2.,1.3])

        # Get the gradient for variables 0,1,2, evaluated at several points
        my_obj.product.grad(1., argnums=[0,1,2])

    """
    def __init__(self, func):
        """Constructor, wraps `func`"""
        self._func = func
        self._obj = None
        self._varnames = func.__code__.co_varnames[:func.__code__.co_argcount]
        self._defaults = func.__defaults__
        self.varidx = {}
        for i, varname in enumerate(self._varnames):
            self.varidx[varname] = i

    def __get__(self, instance, _owner):
        """Hook to connect this wrapper to the instance that has the wrapped member function"""
        self._obj = instance
        return self

    def __call__(self, *args, **kwds):
        """Call the wrapped function"""
        args_ = self.get_args(*args, **kwds)
        return jnp.array([self._func(*args__) for args__ in args_])

    @property
    def func(self):
        """Return the inner function"""
        return self._func

    def get_args(self, *args, **kwds):
        """Get the arguments"""
        #use_args = [arg for arg in args]
        use_args = list(args)
        nargs = len(use_args)
        for varname in self._varnames[nargs:]:
            use_args.append(kwds.get(varname, getattr(self._obj, varname)))
        if len(use_args) != self._func.__code__.co_argcount:
            raise ValueError("Only found %i arguments, need %i" % (len(use_args), self._func.__code__.co_argcount))
        arrays_in = check_inputs(use_args[0], use_args[1:])
        aa = np.broadcast(*arrays_in)
        if not aa.nd:
            return np.expand_dims(arrays_in, 0)
        return aa

    def grad(self, *args, **kwds):
        """Return the gradient"""
        return jax_wrapper(grad, self._func, self.get_args, self.varidx, *args, **kwds)

    def jacfwd(self, *args, **kwds):
        """Return the jacobian"""
        return jax_wrapper(jacfwd, self._func, self.get_args, self.varidx, *args, **kwds)

    def jacrev(self, *args, **kwds):
        """Return the jacobian"""
        return jax_wrapper(jacrev, self._func, self.get_args, self.varidx, *args, **kwds)

    def hess(self, *args, **kwds):
        """Return the hessian"""
        return jax_wrapper(jax_hess, self._func, self.get_args, self.varidx, *args, **kwds)



#!/usr/bin/env python
""" Tools to manage Property object that can be used as fit parameters.
"""
from copy import deepcopy
from collections.abc import Mapping

import numpy as np

from .utils import defaults_decorator
from .property import Property
from .param_holder import ParamHolder


class Parameter(Property):
    """Property sub-class for defining a numerical Parameter.

    This includes value, bounds, error estimates and fixed/free status
    (i.e., for fitting)

    """

    # Better to keep the structure consistent with Property
    defaults = deepcopy(Property.defaults)
    defaults['bounds'] = (np.nan, 'Allowed bounds for value')
    defaults['errors'] = (np.nan, 'Errors on this parameter')
    defaults['free'] = (False, 'Is this property allowed to vary?')
    defaults['scale'] = (1.0, 'Scale to apply for this property')

    # Overwrite the default dtype
    defaults['dtype'] = (ParamHolder, 'Data type')
    defaults['default'] = (np.nan, 'Default value')

    @defaults_decorator(defaults)
    def __init__(self, **kwargs):
        self.bounds = np.nan
        self.errors = np.nan
        self.free = False
        self.scale = 1.
        super(Parameter, self).__init__(**kwargs)

    def __set__(self, obj, value):
        """Set the value in the client object

        Parameter
        ---------
        obj : ...
            The client object
        value : ...
            The value being set

        Rasies
        ------
        TypeError : The input value is the wrong type (i.e., not castable to Darray)

        ValueError : The input values fail the bounds check

        Notes
        -----

        If value is a dict, this will use `Darray(**value)` to construct the managed value
        Otherwise this will use Darray(value, **defaults) to construct the managed value
        """
        par = getattr(obj, self.private_name, None)

        if par is None:
            if isinstance(value, Mapping):
                par = self.dtype(**value)
            else:
                par = self.dtype(value=value, bounds=self.bounds, errors=self.errors, scale=self.scale, free=self.free)
        else:
            par.update(value)

        super(Parameter, self).__set__(obj, par)

    def __get__(self, obj, objtype=None):
        """Get the value from the client object

        Parameter
        ---------
        obj : ...
            The client object

        Return
        ------
        out : ...
            The requested value
        """
        val = getattr(obj, self.private_name).__call__()
        if self.unit is None:
            return val
        return self.unit(val) #pylint: disable=not-callable

    def todict(self, obj):
        """Extract values as an odict """
        return getattr(obj, self.private_name).todict()

    def tostr(self, obj):
        """Extract values as a string"""
        return str(getattr(obj, self.private_name))

    def _cast_type(self, value, obj=None):
        """Hook took override type casting"""
        return value

    def validate_value(self, obj, value): #pylint: disable=unused-argument,no-self-use
        """Validate a value"""
        value.check_bounds()



# -*- coding: utf-8 -*-
# Author: Douglas Creager <dcreager@dcreager.net>
# This file is placed into the public domain.

# Calculates the current version number.  If possible, this is the
# output of “git describe”, modified to conform to the versioning
# scheme that setuptools uses.  If “git describe” returns an error
# (most likely because we're in an unpacked copy of a release tarball,
# rather than in a git working copy), then we fall back on reading the
# contents of the RELEASE-VERSION file.
#
# To use this script, simply import it your setup.py file, and use the
# results of get_git_version() as your package version:
#
# from version import *
#
# setup(
#     version=get_git_version(),
#     .
#     .
#     .
# )
#
# This will automatically update the RELEASE-VERSION file, if
# necessary.  Note that the RELEASE-VERSION file should *not* be
# checked into git; please add it to your top-level .gitignore file.
#
# You'll probably want to distribute the RELEASE-VERSION file in your
# sdist tarballs; to do this, just create a MANIFEST.in file that
# contains the following line:
#
#   include RELEASE-VERSION

import os
import subprocess
from subprocess import check_output

__all__ = ("get_git_version")

_refname = '$Format: %D$'
_tree_hash = '$Format: %t$'
_commit_info = '$Format:%cd by %aN$'
_commit_hash = '$Format: %h$'


def capture_output(cmd, dirname):

    p = subprocess.Popen(cmd,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         cwd=dirname)
    p.stderr.close()

    output = p.stdout.readlines()

    if not output:
        return None
    else:
        return output[0].strip()


def render_pep440(vcs):
    """Convert git release tag into a form that is PEP440 compliant."""

    if vcs is None:
        return None

    tags = vcs.split('-')

    # Bare version number
    if len(tags) == 1:
        return tags[0]
    else:
        return tags[0] + '+' + '.'.join(tags[1:])


def call_git_describe(abbrev=4):

    dirname = os.path.abspath(os.path.dirname(__file__))

    try:
        has_git_tree = capture_output(['git', 'rev-parse',
                                       '--is-inside-work-tree'], dirname)
    except:
        return None

    if not has_git_tree:
        return None

    try:
        line = check_output(['git', 'describe', '--abbrev=%d' % abbrev,
                             '--dirty', '--tags'], cwd=dirname)

        return line.strip().decode('utf-8')

    except:
        return None


def read_release_keywords(keyword):

    refnames = keyword.strip()
    if refnames.startswith("$Format"):
        return None

    refs = set([r.strip() for r in refnames.strip("()").split(",")])
    TAG = "tag: "
    tags = set([r[len(TAG):] for r in refs if r.startswith(TAG)])
    if not tags:
        return None
    return sorted(tags)[-1]


def read_release_version():
    """Read the release version from ``_version.py``."""
    import re
    dirname = os.path.abspath(os.path.dirname(__file__))

    try:
        f = open(os.path.join(dirname, "_version.py"), "rt")
        for line in f.readlines():

            m = re.match("__version__ = '([^']+)'", line)
            if m:
                ver = m.group(1)
                return ver

    except:
        return None

    return None


def write_release_version(version):
    """Write the release version to ``_version.py``."""
    dirname = os.path.abspath(os.path.dirname(__file__))
    f = open(os.path.join(dirname, "_version.py"), "wt")
    f.write("__version__ = '%s'\n" % version)
    f.close()


def get_git_version(abbrev=4):

    # Read in the version that's currently in _version.py.
    release_version = read_release_version()

    # First try to get the current version using “git describe”.
    git_version = call_git_describe(abbrev)
    git_version = render_pep440(git_version)

    # Try to deduce the version from keyword expansion
    keyword_version = read_release_keywords(_refname)
    keyword_version = render_pep440(keyword_version)

    # If that doesn't work, fall back on the value that's in
    # _version.py.
    if git_version is not None:
        version = git_version
    elif release_version is not None:
        version = release_version
    elif keyword_version is not None:
        version = keyword_version
    else:
        version = 'unknown'

    # If we still don't have anything, that's an error.
    if version is None:
        raise ValueError("Cannot find the version number!")

    # If the current version is different from what's in the
    # _version.py file, update the file to be current.
    if version != release_version and version != 'unknown':
        write_release_version(version)

    # Finally, return the current version.
    return version


if __name__ == "__main__":
    print(get_git_version())



""" Property sub-class refering to properties of other objects
"""
import time

from copy import deepcopy

from .property import Property, defaults_decorator

class Ref(Property):
    """Property sub-class refering to properties of other objects
    """
    defaults = deepcopy(Property.defaults)
    defaults.pop('unit')
    defaults.pop('dtype')
    defaults['owner'] = (None, 'Object that owns datum')
    defaults['attr'] = (None, 'Name of attribute in object that owns datum')

    @defaults_decorator(defaults)
    def __init__(self, **kwargs):
        self.owner = None
        self.attr = None
        self.owner_name = None
        self.attr = None
        super(Ref, self).__init__(**kwargs)

    def __set_name__(self, owner, name):
        """Set the name of the privately managed value"""
        super(Ref, self).__set_name__(owner, name)
        self.owner_name = '_' + name + '_owner'
        if self.attr is None:
            self.attr = self.public_name

    def __get__(self, obj, objtype=None):
        """Get the value from the client object

        Parameter
        ---------
        obj : ...
            The client object

        Return
        ------
        val : ...
            The requested value

        """
        owner = getattr(obj, self.owner_name)
        if owner is None:
            return None
        val = getattr(owner, self.attr)
        setattr(obj, self.private_name, val)
        setattr(obj, self.time_name, time.time())
        return val

    def __set__(self, obj, value):
        """Set the pointer to the referenced object

        Parameter
        ---------
        obj : ...
            The client object
        value : ...
            The referenced object
        """        
        setattr(obj, self.owner_name, value)
        _ = self.__get__(obj, self.public_name)
        setattr(obj, self.time_name, time.time())

    def __delete__(self, obj):
        """Set the value to the default value

        This can be useful for sub-classes that use None
        to indicate an un-initialized value.
        """
        owner = getattr(obj, self.owner_name)
        if owner is not None:
            delattr(owner, self.attr)
        super(Ref, self).__delete__(obj)



from setuptools import setup

from python.cfgmdl.version import get_git_version

setup(
    name="cfgmdl",
    version=get_git_version(),
    author="",
    author_email="",
    url = "https://github.com/KIPAC/cfgmdl",
    package_dir={"":"python"},
    packages=["cfgmdl"],
    description="=Tools for configuration parsing and model building",
    long_description=open("README.md").read(),
    package_data={"": ["README.md", "LICENSE"]},
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        ],
    install_requires=["numpy", "jax", "jaxlib", "pyyaml"],
    python_requires='>=3.7',
)



#!/usr/bin/env python
"""
Classes used to describe aspect of Models.

The base class is `Property` which describes any one property of a model,
such as the name, or some other fixed property.

The `Parameter` class describes variable model parameters.

The `Derived` class describes model properies that are derived
from other model properties.

"""
from collections import OrderedDict as odict

import numpy as np

from .configurable import Configurable
from .parameter import Parameter

class Model(Configurable):
    """Configurable model with parameter management

    """
    def __init__(self, **kwargs):
        """ C'tor.  Build from a set of keyword arguments.
        """
        super(Model, self).__init__(**kwargs)
        self._params = self.find_params()

    @classmethod
    def find_params(cls):
        """Find properties that belong to this model"""
        the_classes = cls.mro()
        params = odict()
        for the_class in the_classes:
            for key, val in the_class.__dict__.items():
                if isinstance(val, Parameter):
                    params[key] = val
        return params


    def get_params(self, pnames=None):
        """ Return a list of Parameter objects

        Parameters
        ----------

        pname : list or None
           If a list get the Parameter objects with those names

           If none, get all the Parameter objects

        Returns
        -------

        params : list
            list of Parameters

        """
        l = []
        if pnames is None:
            pnames = self._params.keys()
        for pname in pnames:
            p = self._params[pname]
            if isinstance(p, Parameter):
                l.append(p)
        return l

    def param_values(self, pnames=None):
        """ Return an array with the parameter values

        Parameters
        ----------

        pname : list or None
           If a list, get the values of the `Parameter` objects with those names

           If none, get all values of all the `Parameter` objects

        Returns
        -------

        values : `np.array`
            Parameter values

        """
        l = self.get_params(pnames)
        v = [p.__get__(self) for p in l]
        return np.array(v)


    def param_str(self, pnames=None):
        """ Return an string with the parameter values

        Parameters
        ----------

        pname : list or None
           If a list, get the values of the `Parameter` objects with those names

           If none, get all values of all the `Parameter` objects

        Returns
        -------

        s : `str`
            Parameter value string
        """
        l = self.get_params(pnames)
        s = ""
        for p in l:
            s += "%s : %s\n" % (p.public_name, p.tostr(self))
        return s
